# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring
# pylint: disable=missing-return-doc
# mypy: disable-error-code=method-assign
from __future__ import annotations

import itertools
from unittest.mock import MagicMock, patch

import cirq
import numpy as np
import pandas as pd
import pytest

from supermarq.qcvv import XEB, XEBSample


def test_xeb_init() -> None:
    experiment = XEB()
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    with pytest.raises(
        RuntimeError, match="No samples to retrieve. The experiment has not been run."
    ):
        experiment.samples  # pylint: disable=W0104

    with pytest.raises(RuntimeError, match="No data to retrieve. The experiment has not been run."):
        experiment.circuit_fidelities  # pylint: disable=W0104

    experiment = XEB(two_qubit_gate=cirq.CX)
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CX
    assert experiment.single_qubit_gate_set == [
        cirq.PhasedXZGate(
            z_exponent=z,
            x_exponent=0.5,
            axis_phase_exponent=a,
        )
        for a, z in itertools.product(np.linspace(start=0, stop=7 / 4, num=8), repeat=2)
    ]

    experiment = XEB(single_qubit_gate_set=[cirq.X])
    assert experiment.num_qubits == 2
    assert experiment.two_qubit_gate == cirq.CZ
    assert experiment.single_qubit_gate_set == [cirq.X]


@pytest.fixture
def xeb_experiment() -> XEB:
    return XEB(single_qubit_gate_set=[cirq.X, cirq.Y, cirq.Z])


def test_build_xeb_circuit() -> None:
    with patch("numpy.random.default_rng") as rng:
        xeb_experiment = XEB(single_qubit_gate_set=[cirq.X, cirq.Y, cirq.Z])
        rng.return_value.choice.side_effect = [
            np.array([[cirq.X, cirq.Y], [cirq.Z, cirq.Y], [cirq.Y, cirq.Z]]),
            np.array([[cirq.X, cirq.Z], [cirq.X, cirq.X], [cirq.Y, cirq.Y]]),
        ]
        circuits = xeb_experiment._build_circuits(num_circuits=2, cycle_depths=[2])

    assert len(circuits) == 2

    qbs = xeb_experiment.qubits
    cirq.testing.assert_same_circuits(
        circuits[0].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Z(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[0].data == {"circuit_depth": 5, "num_cycles": 2, "two_qubit_gate": "CZ"}
    cirq.testing.assert_same_circuits(
        circuits[1].circuit,
        cirq.Circuit(
            [
                cirq.X(qbs[0]),
                cirq.Z(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.X(qbs[0]),
                cirq.X(qbs[1]),
                cirq.TaggedOperation(cirq.CZ(*qbs), "no_compile"),
                cirq.Y(qbs[0]),
                cirq.Y(qbs[1]),
                cirq.measure(qbs),
            ]
        ),
    )
    assert circuits[1].data == {"circuit_depth": 5, "num_cycles": 2, "two_qubit_gate": "CZ"}


def test_xeb_analyse_results(xeb_experiment: XEB) -> None:
    xeb_experiment._samples = MagicMock()
    # Choose example data to give perfect fit with fidelity=0.95
    xeb_experiment._raw_data = pd.DataFrame(
        [
            {
                "cycle_depth": 1,
                "circuit_depth": 3,
                "sum_p(x)p(x)": 0.3,
                "sum_p(x)p^(x)": 0.95**1 * 0.3,
            },
            {
                "cycle_depth": 1,
                "circuit_depth": 3,
                "sum_p(x)p(x)": 0.5,
                "sum_p(x)p^(x)": 0.95**1 * 0.5,
            },
            {
                "cycle_depth": 5,
                "circuit_depth": 11,
                "sum_p(x)p(x)": 0.3,
                "sum_p(x)p^(x)": 0.95**5 * 0.3,
            },
            {
                "cycle_depth": 5,
                "circuit_depth": 11,
                "sum_p(x)p(x)": 0.5,
                "sum_p(x)p^(x)": 0.95**5 * 0.5,
            },
            {
                "cycle_depth": 10,
                "circuit_depth": 21,
                "sum_p(x)p(x)": 0.3,
                "sum_p(x)p^(x)": 0.95**10 * 0.3,
            },
            {
                "cycle_depth": 10,
                "circuit_depth": 21,
                "sum_p(x)p(x)": 0.5,
                "sum_p(x)p^(x)": 0.95**10 * 0.5,
            },
        ]
    )
    results = xeb_experiment.analyze_results()

    assert xeb_experiment.results.cycle_fidelity_estimate == pytest.approx(0.95)
    assert xeb_experiment.results.cycle_fidelity_estimate_std == pytest.approx(0.0, abs=1e-8)

    assert results == xeb_experiment.results

    # Call plotting function to test no errors are raised.
    xeb_experiment.plot_results()


def test_xeb_process_probabilities(xeb_experiment: XEB) -> None:
    qubits = cirq.LineQubit.range(2)

    samples = [
        XEBSample(
            raw_circuit=cirq.Circuit(
                [
                    cirq.X(qubits[0]),
                    cirq.X(qubits[1]),
                    cirq.CX(qubits[0], qubits[1]),
                    cirq.X(qubits[0]),
                    cirq.X(qubits[1]),
                    cirq.measure(qubits),
                ]
            ),
            data={"circuit_depth": 3, "num_cycles": 1, "two_qubit_gate": "CX"},
        )
    ]
    samples[0].probabilities = {"00": 0.1, "01": 0.3, "10": 0.4, "11": 0.2}

    with patch("cirq.Simulator") as mock_simulator:
        mock_simulator.return_value.simulate.return_value.final_state_vector = [0.0, 1.0, 0.0, 0.0]
        data = xeb_experiment._process_probabilities(samples)

    expected_data = pd.DataFrame(
        [
            {
                "cycle_depth": 1,
                "circuit_depth": 3,
                "p(00)": 0.0,
                "p(01)": 1.0,
                "p(10)": 0.0,
                "p(11)": 0.0,
                "p^(00)": 0.1,
                "p^(01)": 0.3,
                "p^(10)": 0.4,
                "p^(11)": 0.2,
                "sum_p(x)p(x)": 1.0,
                "sum_p(x)p^(x)": 0.3,
            }
        ]
    )
    pd.testing.assert_frame_equal(expected_data, data)


def test_xeb_process_probabilities_missing_probs(xeb_experiment: XEB) -> None:
    qubits = cirq.LineQubit.range(2)

    samples = [
        XEBSample(
            raw_circuit=cirq.Circuit(
                [
                    cirq.X(qubits[0]),
                    cirq.X(qubits[1]),
                    cirq.CX(qubits[0], qubits[1]),
                    cirq.X(qubits[0]),
                    cirq.X(qubits[1]),
                    cirq.measure(qubits),
                ]
            ),
            data={"circuit_depth": 3, "num_cycles": 1, "two_qubit_gate": "CX"},
        )
    ]

    with pytest.warns(
        UserWarning,
        match=r"1 sample\(s\) are missing `probabilities`. These samples have been omitted.",
    ):
        data = xeb_experiment._process_probabilities(samples)
    pd.testing.assert_frame_equal(data, pd.DataFrame())


def test_xebsample_sum_probs_square_no_values() -> None:
    sample = XEBSample(raw_circuit=cirq.Circuit(), data={})
    with pytest.raises(RuntimeError, match="`target_probabilities` have not yet been initialised"):
        sample.sum_target_probs_square()


def test_xebsample_sum_cross_sample_probs_no_values() -> None:
    sample = XEBSample(raw_circuit=cirq.Circuit(), data={})
    with pytest.raises(RuntimeError, match="`target_probabilities` have not yet been initialised"):
        sample.sum_target_cross_sample_probs()

    sample.target_probabilities = {"example": 0.6}
    with pytest.raises(RuntimeError, match="`sample_probabilities` have not yet been initialised"):
        sample.sum_target_cross_sample_probs()



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for interleaved randomised benchmarking
"""

from __future__ import annotations

import warnings
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Union  # noqa: MDA400

import cirq
import cirq.circuits
import numpy as np
import numpy.typing as npt
import pandas as pd
import scipy
import seaborn as sns
from tqdm.auto import trange
from tqdm.contrib.itertools import product

from supermarq.qcvv.base_experiment import BenchmarkingExperiment, BenchmarkingResults, Sample


####################################################################################################
# Some handy functions for 1 and 2 qubit Clifford operations
####################################################################################################
def _reduce_single_qubit_clifford_seq(
    gate_seq: list[cirq.SingleQubitCliffordGate],
) -> cirq.SingleQubitCliffordGate:
    """Reduces a list of single qubit clifford gates to a single gate.

    Args:
        gate_seq: The list of gates.
    Returns:
        The single reduced gate.
    """
    cur = gate_seq[0]
    for gate in gate_seq[1:]:
        cur = cur.merged_with(gate)
    return cur


def _reduce_clifford_seq(
    gate_seq: list[cirq.CliffordGate],
) -> cirq.CliffordGate:
    """Reduces a list of multi qubit clifford gates to a single gate.

    Args:
        gate_seq: The list of gates.
    Returns:
        The single reduced gate.
    """
    cur = gate_seq[0].clifford_tableau
    for gate in gate_seq[1:]:
        cur = cur.then(gate.clifford_tableau)
    return cirq.CliffordGate.from_clifford_tableau(cur)


####################################################################################################
# The sets `S1`, `S1_X` and `S1_Y` of single qubit Clifford operations are used to generate
# random two qubit Clifford operations. For details see: https://arxiv.org/pdf/1210.7011 &
# https://arxiv.org/pdf/1402.4848.
# The implementation is adapted from:
# https://github.com/quantumlib/Cirq/blob/main/cirq-core/cirq/experiments/qubit_characterizations.py
####################################################################################################
_S1 = [
    # I
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y**0.5 X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**-0.5 Y**-0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, False], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


_S1_X = [
    # X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**0.5, Y**0.5, X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y**-0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([False, True], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


_S1_Y = [
    # Y**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[False], [True]], dtype=np.dtype("bool")),
            zs=np.array([[True], [False]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # X**-0.5 Y**-0.5 X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [False]], dtype=np.dtype("bool")),
            zs=np.array([[True], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
    # Y X**0.5
    cirq.CliffordGate.from_clifford_tableau(
        cirq.CliffordTableau(
            1,
            rs=np.array([True, False], dtype=np.dtype("bool")),
            xs=np.array([[True], [True]], dtype=np.dtype("bool")),
            zs=np.array([[False], [True]], dtype=np.dtype("bool")),
            initial_state=0,
        )
    ),
]


@dataclass(frozen=True)
class IRBResults(BenchmarkingResults):
    """Data structure for the IRB experiment results."""

    rb_decay_coefficient: float
    """Decay coefficient estimate without the interleaving gate."""
    rb_decay_coefficient_std: float
    """Standard deviation of the decay coefficient estimate without the interleaving gate."""
    irb_decay_coefficient: float | None
    """Decay coefficient estimate with the interleaving gate."""
    irb_decay_coefficient_std: float | None
    """Standard deviation of the decay coefficient estimate with the interleaving gate."""
    average_interleaved_gate_error: float | None
    """Estimate of the interleaving gate error."""
    average_interleaved_gate_error_std: float | None
    """Standard deviation of the estimate for the interleaving gate error."""

    experiment_name = "IRB"


@dataclass(frozen=True)
class RBResults(BenchmarkingResults):
    """Data structure for the RB experiment results."""

    rb_decay_coefficient: float
    """Decay coefficient estimate without the interleaving gate."""
    rb_decay_coefficient_std: float
    """Standard deviation of the decay coefficient estimate without the interleaving gate."""
    average_error_per_clifford: float | None
    """Estimate of the average error per Clifford operation."""
    average_error_per_clifford_std: float | None
    """Standard deviation of the the average error per Clifford operation."""

    experiment_name = "RB"


class IRB(BenchmarkingExperiment[Union[IRBResults, RBResults]]):
    r"""Interleaved random benchmarking (IRB) experiment.

    IRB estimates the gate error of specified Clifford gate, :math:`\mathcal{C}^*`.
    This is achieved by first choosing a random sequence, :math:`\{\mathcal{C_i}\}_m`
    of :math:`m` Clifford gates and then using this to generate two circuits. The first
    is generated by appending to this sequence the single gate that corresponds to the
    inverse of the original sequence. The second circuit it obtained by inserting the
    interleaving gate, :math:`\mathcal{C}^*` after each gate in the sequence and then
    again appending the corresponding inverse element of the new circuit. Thus both
    circuits correspond to the identity operation.

    We run both circuits on the specified target and calculate the probability of measuring
    the resulting state in the ground state, :math:`p(0...0)`. This gives the circuit fidelity

    .. math::

        f(m) = 2p(0...0) - 1

    We can then fit an exponential decay :math:`\log(f) \sim m` to this circuit fidelity
    for each circuit, with decay rates :math:`\alpha` and :math:`\tilde{\alpha}` for the circuit
    without and with interleaving respectively. Finally the gate error of the
    specified gate, :math:`\mathcal{C}^*` is estimated as

    .. math::

        e_{\mathcal{C}^*} = \frac{1}{2} \left(1 - \frac{\tilde{\alpha}}{\alpha}\right)

    For more details see: https://arxiv.org/abs/1203.4550
    """

    def __init__(
        self,
        interleaved_gate: cirq.Gate | None = cirq.Z,
        num_qubits: int | None = 1,
        clifford_op_gateset: cirq.CompilationTargetGateset = cirq.CZTargetGateset(),
        *,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """Constructs an IRB experiment.

        Args:
            interleaved_gate: The Clifford gate to measure the gate error of. If None
                then no interleaving is performed and instead vanilla randomized benchmarking is
                performed.
            num_qubits: The number of qubits to experiment on. Must either be 1 or 2 but is ignored
                if a gate is provided - the number of qubits is instead inferred from the gate.
            clifford_op_gateset: The gateset to use when implementing the clifford operations.
                Defaults to the CZ/GR set.
            random_seed: An optional seed to use for randomization.
        """
        if interleaved_gate is not None:
            num_qubits = interleaved_gate.num_qubits()
        if num_qubits not in [1, 2]:
            raise NotImplementedError(
                "IRB experiment is currently only implemented for single or two qubit use."
            )
        super().__init__(num_qubits=num_qubits, random_seed=random_seed)

        if interleaved_gate is not None:
            self.interleaved_gate: cirq.CliffordGate | None = cirq.CliffordGate.from_op_list(
                [interleaved_gate(*self.qubits)], self.qubits
            )
            self._interleaved_op = interleaved_gate(*self.qubits)
            """The operation being interleaved"""
        else:
            self.interleaved_gate = None

        self.clifford_op_gateset = clifford_op_gateset
        """The gateset to use when implementing Clifford operations."""

    def _clifford_gate_to_circuit(
        self,
        clifford: cirq.CliffordGate,
    ) -> cirq.Circuit:
        """Converts a Clifford gate to a circuit using the desired gateset for the experiment.

        Args:
            clifford: The clifford operation to convert.

        Returns:
            A circuit implementing the desired Clifford gate.
        """
        circuit = cirq.Circuit(
            cirq.decompose_clifford_tableau_to_operations(
                self.qubits, clifford.clifford_tableau  # type: ignore[arg-type]
            )
        )
        return cirq.optimize_for_target_gateset(circuit, gateset=self.clifford_op_gateset)

    def random_single_qubit_clifford(self) -> cirq.SingleQubitCliffordGate:
        """Choose a random single qubit clifford gate.

        Returns:
            The random clifford gate.
        """
        Id = cirq.SingleQubitCliffordGate.I
        H = cirq.SingleQubitCliffordGate.H
        S = cirq.SingleQubitCliffordGate.Z_sqrt
        X = cirq.SingleQubitCliffordGate.X
        Y = cirq.SingleQubitCliffordGate.Y
        Z = cirq.SingleQubitCliffordGate.Z

        set_A = np.array(
            [
                Id,
                S,
                H,
                _reduce_single_qubit_clifford_seq([H, S]),
                _reduce_single_qubit_clifford_seq([S, H]),
                _reduce_single_qubit_clifford_seq([H, S, H]),
            ]
        )

        set_B = np.array([Id, X, Y, Z])

        return _reduce_single_qubit_clifford_seq([self._rng.choice(set_A), self._rng.choice(set_B)])

    def random_two_qubit_clifford(self) -> cirq.CliffordGate:
        """Choose a random two qubit clifford gate.

        For algorithm details see https://arxiv.org/pdf/1402.4848 & https://arxiv.org/pdf/1210.7011.

        Returns:
            The random clifford gate.
        """
        qubits = cirq.LineQubit.range(2)
        a = self.random_single_qubit_clifford()
        b = self.random_single_qubit_clifford()
        idx = self._rng.integers(20)
        if idx == 0:
            return cirq.CliffordGate.from_op_list([a(qubits[0]), b(qubits[1])], qubits)
        elif idx == 1:
            return cirq.CliffordGate.from_op_list(
                [
                    a(qubits[0]),
                    b(qubits[1]),
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[0]) ** -0.5,
                    cirq.Y(qubits[1]) ** 0.5,
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[0]) ** 0.5,
                    cirq.Y(qubits[1]) ** -0.5,
                    cirq.CZ(*qubits),
                    cirq.Y(qubits[1]) ** 0.5,
                ],
                qubits,
            )
        elif 2 <= idx <= 10:
            idx_a = int((idx - 2) / 3)
            idx_b = (idx - 2) % 3
            return cirq.CliffordGate.from_op_list(
                [
                    a(qubits[0]),
                    b(qubits[1]),
                    cirq.CZ(*qubits),
                    _S1[idx_a](qubits[0]),
                    _S1_Y[idx_b](qubits[1]),
                ],
                qubits,
            )

        idx_a = int((idx - 11) / 3)
        idx_b = (idx - 11) % 3
        return cirq.CliffordGate.from_op_list(
            [
                a(qubits[0]),
                b(qubits[1]),
                cirq.CZ(*qubits),
                cirq.Y(qubits[0]) ** 0.5,
                cirq.X(qubits[1]) ** -0.5,
                cirq.CZ(*qubits),
                _S1_Y[idx_a](qubits[0]),
                _S1_X[idx_b](qubits[1]),
            ],
            qubits,
        )

    def random_clifford(self) -> cirq.CliffordGate:
        """Returns:
        A random clifford gate with the correct number of qubits for the current experiment.
        """
        if self.num_qubits == 1:
            return self.random_single_qubit_clifford()
        return self.random_two_qubit_clifford()

    def gates_per_clifford(self, samples: int = 500) -> dict[str, float]:
        """Samples a number of random Clifford operations and calculates the average number of
        single and two qubit gates used to implement them. Note this depends on the gateset chosen
        for the experiment.

        Args:
            samples: Number of samples to use. Defaults to 500.

        Returns:
            A dictionary with the average number of one and two qubit gates used.
        """
        sample = [
            self._clifford_gate_to_circuit(self.random_clifford())
            for _ in trange(samples, desc="Sampling Clifford operations")
        ]
        return {
            "single_qubit_gates": np.mean(
                [
                    sum(1 for op in circuit.all_operations() if len(op.qubits) == 1)
                    for circuit in sample
                ]
            ).item(),
            "two_qubit_gates": np.mean(
                [
                    sum(1 for op in circuit.all_operations() if len(op.qubits) == 2)
                    for circuit in sample
                ]
            ).item(),
        }

    def _build_circuits(self, num_circuits: int, cycle_depths: Iterable[int]) -> Sequence[Sample]:
        """Build a list of randomised circuits required for the IRB experiment.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different cycle depths to use during the experiment.

        Returns:
            The list of experiment samples.
        """
        samples = []
        for _, depth in product(range(num_circuits), cycle_depths, desc="Building circuits"):
            base_sequence = [self.random_clifford() for _ in range(depth)]
            rb_sequence = base_sequence + [
                _reduce_clifford_seq(cirq.inverse(base_sequence))  # type: ignore[arg-type]
            ]
            rb_circuit = cirq.Circuit(self._clifford_gate_to_circuit(gate) for gate in rb_sequence)
            samples.append(
                Sample(
                    raw_circuit=rb_circuit + cirq.measure(sorted(self.qubits)),
                    data={
                        "num_cycles": depth,
                        "circuit_depth": len(rb_circuit),
                        "single_qubit_gates": sum(
                            1 for op in rb_circuit.all_operations() if len(op.qubits) == 1
                        ),
                        "two_qubit_gates": sum(
                            1 for op in rb_circuit.all_operations() if len(op.qubits) == 2
                        ),
                        "experiment": "RB",
                    },
                ),
            )
            if self.interleaved_gate is not None:
                # Find final gate
                irb_sequence = [elem for x in base_sequence for elem in (x, self.interleaved_gate)]
                irb_sequence_final_gate = _reduce_clifford_seq(
                    cirq.inverse(irb_sequence)  # type: ignore[arg-type]
                )

                irb_circuit = cirq.Circuit()
                for gate in base_sequence:
                    irb_circuit += self._clifford_gate_to_circuit(gate)
                    irb_circuit += self._interleaved_op.with_tags("no_compile")
                # Add the final inverting gate
                irb_circuit += self._clifford_gate_to_circuit(
                    irb_sequence_final_gate,
                )

                samples.append(
                    Sample(
                        raw_circuit=irb_circuit + cirq.measure(sorted(self.qubits)),
                        data={
                            "num_cycles": depth,
                            "circuit_depth": len(irb_circuit),
                            "single_qubit_gates": sum(
                                1 for op in irb_circuit.all_operations() if len(op.qubits) == 1
                            ),
                            "two_qubit_gates": sum(
                                1 for op in irb_circuit.all_operations() if len(op.qubits) == 2
                            ),
                            "experiment": "IRB",
                        },
                    ),
                )
        return samples

    def _process_probabilities(self, samples: Sequence[Sample]) -> pd.DataFrame:
        """Processes the probabilities generated by sampling the circuits into the data structures
        needed for analyzing the results.

        Args:
            samples: The list of samples to process the results from.

        Returns:
            A data frame of the full results needed to analyse the experiment.
        """

        records = []
        missing_count = 0  # Count the number of samples that do not have probabilities saved
        for sample in samples:
            if sample.probabilities is not None:
                records.append(
                    {
                        "clifford_depth": sample.data["num_cycles"],
                        "circuit_depth": sample.data["circuit_depth"],
                        "experiment": sample.data["experiment"],
                        "single_qubit_gates": sample.data["single_qubit_gates"],
                        "two_qubit_gates": sample.data["two_qubit_gates"],
                        **sample.probabilities,
                    }
                )
            else:
                missing_count += 1

        if missing_count > 0:
            warnings.warn(
                f"{missing_count} sample(s) are missing probabilities. "
                "These samples have been omitted."
            )

        return pd.DataFrame(records)

    def plot_results(self) -> None:
        """Plot the exponential decay of the circuit fidelity with cycle depth."""
        plot = sns.scatterplot(
            data=self.raw_data,
            x="clifford_depth",
            y="0" * self.num_qubits,
            hue="experiment",
        )
        plot.set_xlabel(r"Cycle depth", fontsize=15)
        plot.set_ylabel(r"Survival probability", fontsize=15)
        plot.set_title(r"Exponential decay of survival probability", fontsize=15)

        rb_fit = self._fit_decay()
        xx = np.linspace(0, np.max(self.raw_data.clifford_depth))
        plot.plot(
            xx,
            self.exp_decay(xx, *rb_fit[0]),
            color="tab:blue",
            linestyle="--",
        )
        plot.fill_between(
            xx,
            self.exp_decay(xx, *(rb_fit[0] - rb_fit[1])),
            self.exp_decay(xx, *(rb_fit[0] + rb_fit[1])),
            alpha=0.5,
            color="tab:blue",
        )
        if self.interleaved_gate is not None:
            irb_fit = self._fit_decay("IRB")
            xx = np.linspace(0, np.max(self.raw_data.clifford_depth))
            plot.plot(
                xx,
                self.exp_decay(xx, *irb_fit[0]),
                color="tab:orange",
                linestyle="--",
            )
            plot.fill_between(
                xx,
                self.exp_decay(xx, *(irb_fit[0] - irb_fit[1])),
                self.exp_decay(xx, *(irb_fit[0] + irb_fit[1])),
                alpha=0.5,
                color="tab:orange",
            )

    def analyze_results(self, plot_results: bool = True) -> IRBResults | RBResults:
        """Analyse the experiment results and estimate the interleaved gate error.

        Args:
            plot_results: Whether to generate plots of the results. Defaults to False.

        Returns:
            A named tuple of the final results from the experiment.
        """

        rb_fit = self._fit_decay("RB")
        rb_decay_coefficient, rb_decay_coefficient_std = rb_fit[0][1], rb_fit[1][1]

        if self.interleaved_gate is None:
            self._results = RBResults(
                target="& ".join(self.targets),
                total_circuits=len(self.samples),
                rb_decay_coefficient=rb_decay_coefficient,
                rb_decay_coefficient_std=rb_decay_coefficient_std,
                average_error_per_clifford=(1 - 2**-self.num_qubits) * (1 - rb_decay_coefficient),
                average_error_per_clifford_std=(1 - 2**-self.num_qubits) * rb_decay_coefficient_std,
            )

            if plot_results:
                self.plot_results()

            return self.results

        else:
            irb_fit = self._fit_decay("IRB")
            irb_decay_coefficient, irb_decay_coefficient_std = irb_fit[0][1], irb_fit[1][1]
            interleaved_gate_error = (1 - irb_decay_coefficient / rb_decay_coefficient) * (
                1 - 2**-self.num_qubits
            )

            interleaved_gate_error_std = (
                (1 - 2**-self.num_qubits) / rb_decay_coefficient
            ) * np.sqrt(
                irb_decay_coefficient_std**2
                + irb_decay_coefficient**2 * rb_decay_coefficient_std**2
            )

            self._results = IRBResults(
                target="& ".join(self.targets),
                total_circuits=len(self.samples),
                rb_decay_coefficient=rb_decay_coefficient,
                rb_decay_coefficient_std=rb_decay_coefficient_std,
                irb_decay_coefficient=irb_decay_coefficient,
                irb_decay_coefficient_std=irb_decay_coefficient_std,
                average_interleaved_gate_error=interleaved_gate_error,
                average_interleaved_gate_error_std=interleaved_gate_error_std,
            )

            if plot_results:
                self.plot_results()

            return self.results

    def _fit_decay(
        self, experiment: str = "RB"
    ) -> tuple[np.typing.NDArray[np.float64], np.typing.NDArray[np.float64]]:
        """Fits the exponential decay function to either the RB or IRB results.

        Args:
            experiment: Either `RB` or `IRB` referring to which data to filter by. Defaults to "RB".

        Returns:
            Tuple of the decay fit parameters and their standard deviations.
        """

        xx = self.raw_data.query(f"experiment == '{experiment}'")["clifford_depth"]
        yy = self.raw_data.query(f"experiment == '{experiment}'")["0" * self.num_qubits]

        popt, pcov = scipy.optimize.curve_fit(
            self.exp_decay,
            xx,
            yy,
            p0=(1.0 - 2**-self.num_qubits, 0.99, 2**-self.num_qubits),
            bounds=(0, 1),
            max_nfev=2000,
        )

        return popt, np.sqrt(np.diag(pcov))

    @staticmethod
    def exp_decay(x: npt.ArrayLike, A: float, alpha: float, B: float) -> npt.ArrayLike:
        r"""Exponential decay of the form

        .. math::

            A \alpha^x + B

        Args:
            x: x
            A: Decay constant
            alpha: Decay coefficient
            B: Additive constant

        Returns:
            Exponential decay applied to x.
        """
        return A * (np.asarray(alpha) ** x) + B



from __future__ import annotations

from collections.abc import Iterator

import cirq
import qiskit

from supermarq.benchmark import Benchmark


class BitCode(Benchmark):
    """Creates a circuit for syndrome measurement in a bit-flip error correcting code.

    Args:
        num_data: The number of data qubits.
        num_rounds: The number of measurement rounds.
        bit_state: A list denoting the state to initialize each data qubit to.

    Returns:
        A `cirq.Circuit` for the bit-flip error correcting code.
    """

    def __init__(self, num_data_qubits: int, num_rounds: int, bit_state: list[int]) -> None:
        if len(bit_state) != num_data_qubits:
            raise ValueError("The length of `bit_state` must match the number of data qubits.")
        if not isinstance(bit_state, list):
            raise ValueError("`bit_state` must be a list[int].")
        else:
            if not set(bit_state).issubset({0, 1}):
                raise ValueError("Entries of `bit_state` must be 0, 1 integers.")
        self.num_data_qubits = num_data_qubits
        self.num_rounds = num_rounds
        self.bit_state = bit_state

    def _measurement_round_cirq(
        self, qubits: list[cirq.LineQubit], round_idx: int
    ) -> Iterator[cirq.Operation]:
        """Generates `cirq.Operation`s for a single measurement round.

        Args:
            qubits: Circuit qubits, assuming data on even indices and measurement on odd indices.

        Returns:
            A `cirq.Operation` iterator with the operations for a measurement round.
        """
        ancilla_qubits = qubits[1::2]
        for qq in range(1, len(qubits), 2):
            yield cirq.CX(qubits[qq - 1], qubits[qq])
        for qq in range(1, len(qubits), 2):
            yield cirq.CX(qubits[qq + 1], qubits[qq])
        yield cirq.measure(*ancilla_qubits, key=f"mcm{round_idx}")
        yield from cirq.reset_each(*ancilla_qubits)

    def circuit(self) -> cirq.Circuit:
        """Generates bit code circuit.

        Returns:
            A `cirq.Circuit`.
        """
        num_qubits = 2 * self.num_data_qubits - 1
        qubits = cirq.LineQubit.range(num_qubits)
        circuit = cirq.Circuit()

        # Initialize the data qubits
        for i in range(self.num_data_qubits):
            if self.bit_state[i] == 1:
                circuit.append(cirq.X(qubits[2 * i]))

        # Apply measurement rounds
        circuit.append(self._measurement_round_cirq(qubits, i) for i in range(self.num_rounds))

        circuit.append(cirq.measure(*qubits, key="meas_all"))

        return circuit

    def _get_ideal_dist(self) -> dict[str, float]:
        """Return the ideal probability distribution of `self.circuit()`.

        Since the only allowed initial states for this benchmark are single product states, there
        is a single bitstring that should be measured in the noiseless case.

        Returns:
            Dictionary with measurement results as keys and probabilites as values.
        """
        ancilla_state, final_state = "", ""
        for i in range(self.num_data_qubits - 1):
            # parity checks
            ancilla_state += str((self.bit_state[i] + self.bit_state[i + 1]) % 2)
            final_state += str(self.bit_state[i]) + "0"
        else:
            final_state += str(self.bit_state[-1])

        ideal_bitstring = [ancilla_state] * self.num_rounds + [final_state]
        return {"".join(ideal_bitstring): 1.0}

    def score(self, counts: dict[str, float]) -> float:
        """Compute benchmark score.

        Device performance is given by the Hellinger fidelity between the experimental results and
        the ideal distribution. The ideal is known based on the `bit_state` parameter.

        Args:
            counts: Dictionary containing the measurement counts from running `self.circuit()`.

        Returns:
            A float with the computed score.
        """
        ideal_dist = self._get_ideal_dist()
        total_shots = sum(counts.values())
        experimental_dist = {bitstr: shots / total_shots for bitstr, shots in counts.items()}
        return qiskit.quantum_info.hellinger_fidelity(ideal_dist, experimental_dist)



# Copyright 2021 The Cirq Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# pylint: disable=missing-function-docstring
# pylint: disable=missing-return-doc
# mypy: disable-error-code=method-assign

from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import MagicMock, call, patch

import cirq
import cirq_superstaq as css
import pandas as pd
import pytest

from supermarq.qcvv.base_experiment import BenchmarkingExperiment, BenchmarkingResults, Sample


@dataclass(frozen=True)
class ExampleResults(BenchmarkingResults):
    """NamedTuple instance to use for testing"""

    example: float

    experiment_name = "Example results"


@pytest.fixture
@patch.multiple(BenchmarkingExperiment, __abstractmethods__=set())
def abc_experiment() -> BenchmarkingExperiment[ExampleResults]:
    return BenchmarkingExperiment(num_qubits=2)  # type: ignore[abstract]


@pytest.fixture
def sample_circuits() -> list[Sample]:
    qubits = cirq.LineQubit.range(2)
    return [
        Sample(
            raw_circuit=cirq.Circuit(cirq.CZ(*qubits), cirq.CZ(*qubits), cirq.measure(*qubits)),
            data={"circuit": 1},
        ),
        Sample(
            raw_circuit=cirq.Circuit(cirq.CX(*qubits), cirq.measure(*qubits)), data={"circuit": 2}
        ),
    ]


def test_sample_target_property() -> None:
    sample = Sample(raw_circuit=MagicMock(), data={})
    assert sample.target == "No target"

    sample.probabilities = {"0": 0.25, "1": 0.75}
    assert sample.target == "Local simulator"

    sample.job = MagicMock()
    sample.job.target.return_value = "Example target"
    assert sample.target == "Example target"


def test_benchmarking_experiment_init(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
) -> None:
    assert abc_experiment.num_qubits == 2
    assert abc_experiment._raw_data is None
    assert abc_experiment._results is None
    assert abc_experiment._samples is None

    abc_experiment._raw_data = pd.DataFrame([{"Example": 0.1}])
    abc_experiment._results = ExampleResults(target="Some target", total_circuits=2, example=5.0)

    pd.testing.assert_frame_equal(abc_experiment.raw_data, abc_experiment._raw_data)
    assert abc_experiment.results == abc_experiment._results


def test_empty_results_error(abc_experiment: BenchmarkingExperiment[ExampleResults]) -> None:
    with pytest.raises(
        RuntimeError, match="No results to retrieve. The experiment has not been run."
    ):
        _ = abc_experiment.results


def test_empty_data_error(abc_experiment: BenchmarkingExperiment[ExampleResults]) -> None:
    with pytest.raises(RuntimeError, match="No data to retrieve. The experiment has not been run."):
        _ = abc_experiment.raw_data


def test_empty_samples_error(abc_experiment: BenchmarkingExperiment[ExampleResults]) -> None:
    with pytest.raises(
        RuntimeError, match="No samples to retrieve. The experiment has not been run."
    ):
        _ = abc_experiment.samples


def test_prepare_experiment_overwrite_error(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
) -> None:
    abc_experiment._samples = [Sample(raw_circuit=MagicMock(), data={})]
    abc_experiment._build_circuits = MagicMock()

    with pytest.raises(
        RuntimeError,
        match="This experiment already has existing data which would be overwritten by "
        "rerunning the experiment. If this is the desired behavior set `overwrite=True`",
    ):
        abc_experiment.prepare_experiment(100, [1, 50, 100])


def test_prepare_experiment_overwrite(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
) -> None:
    abc_experiment._samples = [Sample(raw_circuit=MagicMock(), data={})]
    abc_experiment._build_circuits = MagicMock()
    abc_experiment._validate_circuits = MagicMock()

    abc_experiment.prepare_experiment(100, [1, 50, 100], overwrite=True)

    abc_experiment._build_circuits.assert_called_once_with(100, [1, 50, 100])
    abc_experiment._validate_circuits.assert_called_once_with()


def test_prepare_experiment_with_bad_layers(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
) -> None:
    with pytest.raises(
        ValueError, match="The `cycle_depths` iterator can only include positive values."
    ):
        abc_experiment.prepare_experiment(20, [0])


def test_run_with_simulator(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    cirq.measurement_key_name = MagicMock()
    abc_experiment._samples = sample_circuits
    test_sim = MagicMock()
    mock_result = MagicMock()
    mock_result.histogram.return_value = {0: 0, 1: 100, 2: 0, 3: 0}
    test_sim.run.return_value = mock_result

    abc_experiment.run_with_simulator(simulator=test_sim, repetitions=100)

    # Test simulator calls
    test_sim.run.assert_has_calls(
        [
            call(sample_circuits[0].circuit, repetitions=100),
            call(sample_circuits[1].circuit, repetitions=100),
        ],
        any_order=True,
    )

    # Test probabilities
    assert sample_circuits[0].probabilities == {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}
    assert sample_circuits[1].probabilities == {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}


def test_run_with_simulator_existing_probabilties(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    sample_circuits[0].probabilities = {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}
    sample_circuits[1].probabilities = {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}

    abc_experiment._samples = sample_circuits
    with pytest.raises(
        RuntimeError,
        match=(
            "Some samples have already been run. Re-running the experiment will"
            "overwrite these results. If this is the desired behaviour use `overwrite=True`"
        ),
    ):
        abc_experiment.run_with_simulator()


def test_run_with_simulator_default_target(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    cirq.measurement_key_name = MagicMock()
    cirq.Simulator = (target := MagicMock())  # type: ignore [misc]
    abc_experiment._samples = sample_circuits
    mock_result = MagicMock()
    mock_result.histogram.return_value = {0: 0, 1: 100, 2: 0, 3: 0}
    target().run.return_value = mock_result

    abc_experiment.run_with_simulator(repetitions=100)

    # Test simulator calls
    target().run.assert_has_calls(
        [
            call(sample_circuits[0].circuit, repetitions=100),
            call(sample_circuits[1].circuit, repetitions=100),
        ],
        any_order=True,
    )

    # Test probabilities
    assert sample_circuits[0].probabilities == {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}
    assert sample_circuits[1].probabilities == {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}


def test_run_on_device(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        job = abc_experiment.run_on_device(
            target="example_target", repetitions=100, overwrite=False, **{"some": "options"}
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].raw_circuit, sample_circuits[1].raw_circuit],
        target="example_target",
        method=None,
        repetitions=100,
        some="options",
    )

    assert job == mock_service.return_value.create_job.return_value
    assert (
        sample_circuits[0].compiled_circuit
        == mock_service.return_value.create_job.return_value.compiled_circuits.return_value[0]
    )
    assert (
        sample_circuits[1].compiled_circuit
        == mock_service.return_value.create_job.return_value.compiled_circuits.return_value[1]
    )


def test_run_on_device_existing_probabilties(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    sample_circuits[0].probabilities = {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}
    sample_circuits[1].probabilities = {"00": 0.0, "01": 1.0, "10": 0.0, "11": 0.0}

    abc_experiment._samples = sample_circuits
    with pytest.raises(
        RuntimeError,
        match=(
            "Some samples have already been run. Re-running the experiment will"
            "overwrite these results. If this is the desired behaviour use `overwrite=True`"
        ),
    ):
        abc_experiment.run_on_device(target="example")


def test_run_on_device_dry_run(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        job = abc_experiment.run_on_device(
            target="example_target", repetitions=100, method="dry-run"
        )

    mock_service.return_value.create_job.assert_called_once_with(
        [sample_circuits[0].raw_circuit, sample_circuits[1].raw_circuit],
        target="example_target",
        method="dry-run",
        repetitions=100,
    )
    assert job == mock_service.return_value.create_job.return_value
    assert (
        sample_circuits[0].compiled_circuit
        == mock_service.return_value.create_job.return_value.compiled_circuits.return_value[0]
    )
    assert (
        sample_circuits[1].compiled_circuit
        == mock_service.return_value.create_job.return_value.compiled_circuits.return_value[1]
    )


def test_interleave_circuit() -> None:
    qubit = cirq.LineQubit(0)
    circuit = cirq.Circuit(*[cirq.X(qubit) for _ in range(4)])

    # With last gate
    interleaved_circuit = BenchmarkingExperiment._interleave_op(
        circuit, cirq.Z(qubit), include_final=True
    )
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
        ),
    )

    # Without last gate
    interleaved_circuit = BenchmarkingExperiment._interleave_op(
        circuit, cirq.Z(qubit), include_final=False
    )
    cirq.testing.assert_same_circuits(
        interleaved_circuit,
        cirq.Circuit(
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
            cirq.TaggedOperation(cirq.Z(qubit), "no_compile"),
            cirq.X(qubit),
        ),
    )


def test_sample_statuses(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits
    mock_job = MagicMock(spec=css.Job)
    mock_job.status.return_value = "example_status"
    sample_circuits[0].job = mock_job

    statuses = abc_experiment._sample_statuses()
    assert statuses == ["example_status", None]
    mock_job.status.assert_called_once_with()


def test_targets(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
) -> None:
    abc_experiment._samples = [mock_sample_0 := MagicMock(), mock_sample_1 := MagicMock()]
    mock_sample_0.target = "target_0"
    mock_sample_1.target = "target_1"
    assert abc_experiment.targets == frozenset(["target_0", "target_1"])

    mock_sample_1.target = "target_0"
    assert abc_experiment.targets == frozenset(["target_0"])


def test_validate_circuits(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits
    # Should't get any errors with the base circuits
    abc_experiment._validate_circuits()

    # Add a gate so not all measurements are terminal
    abc_experiment._samples[0].raw_circuit += cirq.X(abc_experiment.qubits[0])
    with pytest.raises(
        ValueError, match="QCVV experiment circuits can only contain terminal measurements."
    ):
        abc_experiment._validate_circuits()

    # Remove measurements
    abc_experiment._samples[0].raw_circuit = abc_experiment._samples[0].circuit[:-2] + cirq.measure(
        abc_experiment.qubits[0]
    )
    with pytest.raises(
        ValueError,
        match="The terminal measurement in QCVV experiment circuits must measure all qubits.",
    ):
        abc_experiment._validate_circuits()

    # Remove all measurements
    abc_experiment._samples[0].raw_circuit = abc_experiment._samples[0].circuit[:-2]
    with pytest.raises(
        ValueError,
        match="QCVV experiment circuits must contain measurements.",
    ):
        abc_experiment._validate_circuits()


def test_process_device_counts(abc_experiment: BenchmarkingExperiment[ExampleResults]) -> None:
    counts = {
        "00": 20,
        "01": 5,
        "11": 10,
    }
    probs = abc_experiment._process_device_counts(counts)

    assert probs == {"00": 20 / 35, "01": 5 / 35, "10": 0.0, "11": 10 / 35}


def test_retrieve_jobs_not_all_submitted(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits

    mock_job_1 = MagicMock()
    mock_job_1.status.return_value = "Queued"
    mock_job_1.job_id.return_value = "example_job_id"
    abc_experiment._samples[0].job = mock_job_1

    statuses = abc_experiment._retrieve_jobs()

    assert statuses == {"example_job_id": "Queued"}
    assert sample_circuits[0].probabilities is None
    assert sample_circuits[1].probabilities is None


def test_retrieve_jobs_nothing_to_retrieve(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits
    statuses = abc_experiment._retrieve_jobs()
    assert statuses == {}


def test_retrieve_jobs_all_submitted(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits
    mock_job_1 = MagicMock(spec=css.Job)
    mock_job_1.status.return_value = "Queued"
    mock_job_1.job_id.return_value = "example_job_id_1"
    abc_experiment._samples[0].job = mock_job_1

    mock_job_2 = MagicMock(spec=css.Job)
    mock_job_2.status.return_value = "Done"
    mock_job_2.job_id.return_value = "example_job_id_2"
    mock_job_2.counts.return_value = {"00": 5, "11": 10}
    abc_experiment._samples[1].job = mock_job_2

    statuses = abc_experiment._retrieve_jobs()

    # Check counts call
    mock_job_2.counts.assert_called_once_with(0)

    assert statuses == {"example_job_id_1": "Queued"}

    # Check probabilities correctly updated
    assert sample_circuits[1].probabilities == {"00": 5 / 15, "01": 0.0, "10": 0.0, "11": 10 / 15}
    assert sample_circuits[0].probabilities is None


def test_collect_data_no_samples(abc_experiment: BenchmarkingExperiment[ExampleResults]) -> None:
    with pytest.raises(RuntimeError, match="The experiment has not yet ben run."):
        abc_experiment.collect_data()


def test_collect_data_no_jobs_to_retrieve(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    sample_circuits[0].probabilities = {"00": 1.0, "10": 0.0, "01": 0.0, "11": 0.0}
    sample_circuits[1].probabilities = {"00": 0.0, "10": 0.0, "01": 0.0, "11": 1.0}
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock()

    assert abc_experiment.collect_data()
    abc_experiment._process_probabilities.assert_called_once_with(sample_circuits)


def test_collect_data_no_jobs_to_retrieve_not_all_probabilities(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
    capfd: pytest.CaptureFixture[str],
) -> None:
    sample_circuits[0].probabilities = {"00": 1.0, "10": 0.0, "01": 0.0, "11": 0.0}
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock()

    assert not abc_experiment.collect_data()
    out, _ = capfd.readouterr()
    assert out == "Some samples do not have probability results.\n"
    abc_experiment._process_probabilities.assert_not_called()


def test_collect_data_no_jobs_to_retrieve_not_all_probabilities_forced(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    sample_circuits[0].probabilities = {"00": 1.0, "10": 0.0, "01": 0.0, "11": 0.0}
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock()

    assert abc_experiment.collect_data(force=True)
    abc_experiment._process_probabilities.assert_called_once_with([sample_circuits[0]])


def test_collect_data_cannot_force(
    abc_experiment: BenchmarkingExperiment[ExampleResults], sample_circuits: list[Sample]
) -> None:
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock()

    with pytest.raises(
        RuntimeError, match="Cannot force data collection when there are no completed samples."
    ):
        abc_experiment.collect_data(force=True)

    abc_experiment._process_probabilities.assert_not_called()


def test_collect_data_outstanding_jobs(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
    capfd: pytest.CaptureFixture[str],
) -> None:
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock()
    abc_experiment._retrieve_jobs = MagicMock(return_value={"example_id": "some_status"})
    assert not abc_experiment.collect_data()
    out, _ = capfd.readouterr()
    assert out == (
        "Not all circuits have been sampled. Please wait and try again.\n"
        "Outstanding Superstaq jobs:\n"
        "{'example_id': 'some_status'}\n"
    )
    abc_experiment._process_probabilities.assert_not_called()


def test_collect_data_outstanding_jobs_force(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
    capfd: pytest.CaptureFixture[str],
) -> None:
    abc_experiment._samples = sample_circuits
    abc_experiment._process_probabilities = MagicMock(return_value=pd.DataFrame([{"data": 1.0}]))
    abc_experiment.samples[0].probabilities = {"00": 1.0, "10": 0.0, "01": 0.0, "11": 0.0}
    abc_experiment._retrieve_jobs = MagicMock(return_value={"example_id": "some_status"})
    assert abc_experiment.collect_data(force=True)
    out, _ = capfd.readouterr()
    assert out == (
        "Not all circuits have been sampled. Please wait and try again.\n"
        "Outstanding Superstaq jobs:\n"
        "{'example_id': 'some_status'}\n"
        "Some samples do not have probability results.\n"
    )

    abc_experiment._process_probabilities.assert_called_once_with([abc_experiment.samples[0]])

    pd.testing.assert_frame_equal(pd.DataFrame([{"data": 1.0}]), abc_experiment.raw_data)


def test_compile_circuit(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
) -> None:
    abc_experiment._samples = sample_circuits

    with patch("cirq_superstaq.Service") as mock_service:
        mock_service.return_value.compile.return_value.circuits = (
            mock_compiled_circuits := MagicMock()
        )
        abc_experiment.compile_circuits("example_target", additional="kwargs")

    mock_service.return_value.compile.assert_called_once_with(
        [sample_circuits[0].raw_circuit, sample_circuits[1].raw_circuit],
        target="example_target",
        additional="kwargs",
    )
    assert sample_circuits[0].circuit == mock_compiled_circuits[0]
    assert sample_circuits[1].circuit == mock_compiled_circuits[1]


def test_run_with_callable(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
) -> None:
    abc_experiment._samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"01": 0.2, "10": 0.7, "11": 0.1}

    abc_experiment.run_with_callable(test_callable, some="kwargs")

    test_callable.assert_has_calls(
        [
            call(sample_circuits[0].circuit, some="kwargs"),
            call(sample_circuits[1].circuit, some="kwargs"),
        ]
    )
    assert sample_circuits[0].probabilities == {"00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1}
    assert sample_circuits[1].probabilities == {"00": 0.0, "01": 0.2, "10": 0.7, "11": 0.1}


def test_run_with_callable_bad_bitstring(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
) -> None:
    abc_experiment._samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"000": 0.0, "01": 0.2, "10": 0.8}

    with pytest.raises(
        RuntimeError, match="Returned probabilities include an incorrect number of bits."
    ):
        abc_experiment.run_with_callable(test_callable, some="kwargs")


def test_run_with_callable_bad_probabilities(
    abc_experiment: BenchmarkingExperiment[ExampleResults],
    sample_circuits: list[Sample],
) -> None:
    abc_experiment._samples = sample_circuits
    test_callable = MagicMock()
    test_callable.return_value = {"00": 0.0, "01": 0.2, "10": 0.7, "11": 0.09}

    with pytest.raises(RuntimeError, match="Returned probabilities do not sum to 1.0."):
        abc_experiment.run_with_callable(test_callable, some="kwargs")



# pylint: disable=missing-function-docstring,missing-class-docstring
import packaging.version

import supermarq


def test_version() -> None:
    assert (
        packaging.version.Version("0.0.4")
        < packaging.version.parse(supermarq.__version__)
        < packaging.version.Version("1.0.0")
    )



"""Definition of the Fermionic SWAP QAOA benchmark within the Supermarq suite."""

from __future__ import annotations

import cirq
import numpy as np

from supermarq.benchmarks.qaoa_vanilla_proxy import QAOAVanillaProxy


class QAOAFermionicSwapProxy(QAOAVanillaProxy):
    """Proxy of a full Quantum Approximate Optimization Algorithm (QAOA) benchmark.

    This benchmark targets MaxCut on a Sherrington-Kirkpatrick (SK) model. Device
    performance is given by the Hellinger fidelity between the experimental output
    distribution and the true distribution obtained via scalable, classical simulation.

    The ansatz for this QAOA problem utilizes the fermionic SWAP network which is able
    to perform all of the required O(N^2) interactions in linear circuit depth. This
    ansatz is especially well-suited to QPU architectures which only support
    nearest-neighbor connectivity. See https://doi.org/10.3390/electronics10141690 for
    an example of this ansatz used in practice.

    When a new instance of this benchmark is created, the ansatz parameters will
    be initialized by:

    #. Generating a random instance of an SK graph

    #. Finding approximately optimal angles (rather than random values)
    """

    def _gen_ansatz(self, gamma: float, beta: float) -> cirq.Circuit:
        qubits = cirq.LineQubit.range(self.num_qubits)
        circuit = cirq.Circuit()

        # initialize |++++>
        for q in qubits:
            circuit.append(cirq.H(q))

        # Implement the phase-separator unitary with a swap network
        # The covers indicate which qubits will be swapped at each layer
        cover_a = [(idx - 1, idx) for idx in range(1, self.num_qubits, 2)]
        cover_b = [(idx - 1, idx) for idx in range(2, self.num_qubits, 2)]

        # The indices of the virtual map correspond to physical qubits,
        # the value at that index corresponds to the virtual qubit residing there
        virtual_map = np.arange(self.num_qubits)

        for layer in range(self.num_qubits):
            cover = [cover_a, cover_b][layer % 2]
            for pair in cover:
                i, j = pair  # swap physical qubits i and j

                # Get the corresponding weight between the virtual qubits
                v_i = virtual_map[i]
                v_j = virtual_map[j]
                for edge in self.hamiltonian:
                    if v_i == edge[0] and v_j == edge[1]:
                        weight = edge[2]
                phi = gamma * weight

                # Perform the ZZ+SWAP operation
                circuit.append(cirq.CNOT(qubits[i], qubits[j]))
                circuit.append(cirq.rz(2 * phi)(qubits[j]))
                circuit.append(cirq.CNOT(qubits[j], qubits[i]))
                circuit.append(cirq.CNOT(qubits[i], qubits[j]))

                # update the virtual map
                virtual_map[j], virtual_map[i] = virtual_map[i], virtual_map[j]

        # Implement the mixing unitary
        for q in qubits:
            circuit.append(cirq.rx(2 * beta)(q))

        # Measure all qubits
        circuit.append(cirq.measure(*qubits))

        # NOTE: the final qubits in this circuit are in REVERSED order due to the swap network
        return circuit

    def _get_energy_for_bitstring(self, bitstring: str) -> float:
        # Reverse bitstring ordering due to SWAP network
        return super()._get_energy_for_bitstring(bitstring[::-1])



from __future__ import annotations

from collections.abc import Mapping

import cirq
import numpy as np
import numpy.typing as npt
import scipy

import supermarq
from supermarq.benchmark import Benchmark


class QAOAVanillaProxy(Benchmark):
    """Proxy of a full Quantum Approximate Optimization Algorithm (QAOA) benchmark.

    This benchmark targets MaxCut on a Sherrington-Kirkpatrick (SK) model.
    Device performance is given by the Hellinger fidelity between the experimental
    output distribution and the true distribution obtained via scalable, classical
    simulation.

    The ansatz for this QAOA problem follows the typical structure obtained when directly
    translating the objective Hamiltonian to the quantum circuit. Since the SK model is
    completely connected there are O(N^2) interactions that need to take place. These
    are implementation by pairs of CNOTs and Rz rotations between the participating qubits.
    This ansatz is well-suited to QPU architectures which support all-to-all connectivity.

    When a new instance of this benchmark is created, the ansatz parameters will
    be initialized by:

    #. Generating a random instance of an SK graph

    #. Finding approximately optimal angles (rather than random values)
    """

    def __init__(self, num_qubits: int) -> None:
        """Generate a new benchmark instance.

        Args:
            num_qubits: The number of nodes (qubits) within the SK graph.
        """
        self.num_qubits = num_qubits
        self.hamiltonian = self._gen_sk_hamiltonian()
        self.params = self._gen_angles()

    def _gen_sk_hamiltonian(self) -> list[tuple[int, int, float]]:
        """Randomly pick +1 or -1 for each edge weight."""
        hamiltonian = []
        for i in range(self.num_qubits):
            for j in range(i + 1, self.num_qubits):
                hamiltonian.append((i, j, np.random.choice([-1, 1])))

        np.random.shuffle(hamiltonian)

        return hamiltonian

    def _gen_ansatz(self, gamma: float, beta: float) -> cirq.Circuit:
        qubits = cirq.LineQubit.range(self.num_qubits)
        circuit = cirq.Circuit()

        # initialize |++++>
        for q in qubits:
            circuit.append(cirq.H(q))

        # Apply the phase separator unitary
        for term in self.hamiltonian:
            i, j, weight = term
            phi = gamma * weight

            # Perform a ZZ interaction
            circuit.append(cirq.CNOT(qubits[i], qubits[j]))
            circuit.append(cirq.rz(2 * phi)(qubits[j]))
            circuit.append(cirq.CNOT(qubits[i], qubits[j]))

        # Apply the mixing unitary
        for q in qubits:
            circuit.append(cirq.rx(2 * beta)(q))

        # Measure all qubits
        circuit.append(cirq.measure(*qubits))

        return circuit

    def _get_energy_for_bitstring(self, bitstring: str) -> float:
        energy = 0.0
        for i, j, weight in self.hamiltonian:
            if bitstring[i] == bitstring[j]:
                energy -= weight  # if edge is UNCUT, weight counts against objective
            else:
                energy += weight  # if edge is CUT, weight counts towards objective
        return energy

    def _get_expectation_value_from_probs(self, probabilities: Mapping[str, float]) -> float:
        expectation_value = 0.0
        for bitstring, probability in probabilities.items():
            expectation_value += probability * self._get_energy_for_bitstring(bitstring)
        return expectation_value

    def _get_opt_angles(self) -> tuple[npt.NDArray[np.float64], float]:
        def f(params: npt.NDArray[np.float64]) -> float:
            """The objective function to minimize.

            Args:
                params: The parameters at which to evaluate the objective.

            Returns:
                Evaluation of objective given parameters.
            """
            gamma, beta = params
            circ = self._gen_ansatz(gamma, beta)
            probs = supermarq.simulation.get_ideal_counts(circ)
            objective_value = self._get_expectation_value_from_probs(probs)

            return -objective_value  # because we are minimizing instead of maximizing

        init_params = [np.random.uniform() * 2 * np.pi, np.random.uniform() * 2 * np.pi]
        out = scipy.optimize.minimize(f, init_params, method="COBYLA")

        return out["x"], out["fun"]

    def _gen_angles(self) -> npt.NDArray[np.float64]:
        # Classically simulate the variational optimization 5 times,
        # return the parameters from the best performing simulation
        best_params, best_cost = np.zeros(2), 0.0
        for _ in range(5):
            params, cost = self._get_opt_angles()
            if cost < best_cost:
                best_params = params
                best_cost = cost
        return best_params

    def circuit(self) -> cirq.Circuit:
        """Generate a QAOA circuit for the Sherrington-Kirkpatrick model.

        The ansatz structure is given by the form of the Hamiltonian and requires interactions
        between every pair of qubits. We restrict the depth of this proxy benchmark to p=1 to keep
        the classical simulation scalable.

        Returns:
            The S-K model QAOA `cirq.Circuit`.
        """
        gamma, beta = self.params
        return self._gen_ansatz(gamma, beta)

    def score(self, counts: Mapping[str, float]) -> float:
        """Compare the experimental output to the output of noiseless simulation.

        The implementation here has exponential runtime and would not scale. However, it could in
        principle be done efficiently via https://arxiv.org/abs/1706.02998, so we're good.

        Args:
            counts: A dictionary containing the measurement counts from circuit execution.

        Returns:
            The QAOA proxy benchmark score.
        """
        ideal_counts = supermarq.simulation.get_ideal_counts(self.circuit())
        total_shots = sum(counts.values())
        experimental_counts = {k: v / total_shots for k, v in counts.items()}

        ideal_value = self._get_expectation_value_from_probs(ideal_counts)
        experimental_value = self._get_expectation_value_from_probs(experimental_counts)

        return 1 - abs(ideal_value - experimental_value) / (2 * ideal_value)



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tooling for cross entropy benchmark experiments.
"""
from __future__ import annotations

import itertools
import warnings
from collections.abc import Iterable, Sequence
from dataclasses import dataclass

import cirq
import numpy as np
import pandas as pd
import scipy
import seaborn as sns
import tqdm.auto
import tqdm.contrib.itertools

from supermarq.qcvv.base_experiment import BenchmarkingExperiment, BenchmarkingResults, Sample


@dataclass
class XEBSample(Sample):
    """The samples used in XEB experiments."""

    target_probabilities: dict[str, float] | None = None
    """The target probabilities obtained through a noiseless simulator"""
    sample_probabilities: dict[str, float] | None = None
    """The sample probabilities obtained from the chosen target"""

    def sum_target_probs_square(self) -> float:
        """Compute the sum of the squared target probabilities.

        Raises:
            RuntimeError: If no target probabilities have been initialised.

        Returns:
            float: The sum of squared target probabilities.
        """
        if self.target_probabilities is None:
            raise RuntimeError("`target_probabilities` have not yet been initialised")

        return sum(prob**2 for prob in self.target_probabilities.values())

    def sum_target_cross_sample_probs(self) -> float:
        """Compute the dot product between the sample and target probabilities

        Raises:
            RuntimeError: If either the target or sample probabilities have not yet been
                initialised.

        Returns:
            float: The dot product between the sample and target probabilities.
        """
        if self.target_probabilities is None:
            raise RuntimeError("`target_probabilities` have not yet been initialised")

        if self.sample_probabilities is None:
            raise RuntimeError("`sample_probabilities` have not yet been initialised")

        return sum(
            self.target_probabilities[state] * self.sample_probabilities[state]
            for state in self.target_probabilities
        )


@dataclass(frozen=True)
class XEBResults(BenchmarkingResults):
    """Results from an XEB experiment."""

    cycle_fidelity_estimate: float
    """Estimated cycle fidelity."""
    cycle_fidelity_estimate_std: float
    """Standard deviation for the cycle fidelity estimate."""

    experiment_name = "XEB"


class XEB(BenchmarkingExperiment[XEBResults]):
    r"""Cross-entropy benchmarking (XEB) experiment.

    The XEB experiment can be used to estimate the combined fidelity of a repeating
    cycle of gates. In our case, where we restrict ourselves to two qubits, we use
    cycles made up of two randomly selected single qubit phased XZ gates and a constant
    two qubit gate. This is illustrated as follows:

    For each randomly generated circuit, with a given number of cycle, we compare the
    simulated state probabilities, :math:`p(x)` with those achieved by running the circuit
    on a given target, :math:`\hat{p}(x)`. The fidelity of a circuit containing :math:`d`
    cycles, :math:`f_d` can then be estimated as

    .. math::

        \sum_{x \in \{0, 1\}^n} p(x) \hat{p}(x) - \frac{1}{2^n} =
        f_d \left(\sum_{x \in \{0, 1\}^n} p(x)^2 -  \frac{1}{2^n}\right)

    We can therefore fit a linear model to estimate the value of :math:`f_d`. We the estimate
    the fidelity of the cycle, :math:`f_{\mathrm{cycle}}` as

    .. math::

        f_d = A(f_{cycle})^d

    Thus fitting another linear model to :math:`\log(f_d) \sim d` provides us with an estimate
    of the cycle fidelity.

    For more details see: https://www.nature.com/articles/s41586-019-1666-5
    """

    def __init__(
        self,
        single_qubit_gate_set: list[cirq.Gate] | None = None,
        two_qubit_gate: cirq.Gate | None = cirq.CZ,
        *,
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """Initializes a cross-entropy benchmarking experiment.

        Args:
            single_qubit_gate_set: Optional list of single qubit gates to randomly sample from when
                generating random circuits. If not provided defaults to phased XZ gates with 1/4 pi
                intervals.
            two_qubit_gate: The two qubit gate to interleave between the single qubit gates. If None
                then no two qubit gate is used. Defaults to control-Z gate.
            random_seed: An optional seed to use for randomization.
        """
        super().__init__(num_qubits=2, random_seed=random_seed)

        self._circuit_fidelities: pd.DataFrame | None = None

        self._samples: Sequence[XEBSample] | None = None  # Overwrite with modified sampled object

        self.two_qubit_gate: cirq.Gate | None = two_qubit_gate
        """The two qubit gate to use for interleaving."""

        self.single_qubit_gate_set: list[cirq.Gate]
        """The single qubit gates to randomly sample from"""

        if single_qubit_gate_set is None:
            gate_exponents = np.linspace(start=0, stop=7 / 4, num=8)
            self.single_qubit_gate_set = [
                cirq.PhasedXZGate(
                    z_exponent=z,  # 1) Choose an axis in the xy-plane, zπ from the +x-axis.
                    x_exponent=0.5,  # 2) Rotate about the axis in 1) by a fixed π/2.
                    axis_phase_exponent=a,  # 3) Rotate about the +z-axis by aπ (a final phasing).
                )
                for a, z in itertools.product(
                    gate_exponents, repeat=2
                )  # enumerates every possible (a, z)
            ]
        else:
            self.single_qubit_gate_set = single_qubit_gate_set

    ##############
    # Properties #
    ##############
    @property
    def circuit_fidelities(self) -> pd.DataFrame:
        """The circuit fidelity calculations from the most recently run experiment.

        Raises:
            RuntimeError: If no data is available.
        """
        if self._circuit_fidelities is None:
            raise RuntimeError("No data to retrieve. The experiment has not been run.")

        return self._circuit_fidelities

    @property
    def samples(self) -> Sequence[XEBSample]:  # Overwrite with XEBSample return type
        """The samples generated during the experiment.

        Raises:
            RuntimeError: If no samples are available.
        """
        if self._samples is None:
            raise RuntimeError("No samples to retrieve. The experiment has not been run.")

        return self._samples

    ###################
    # Private Methods #
    ###################
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[XEBSample]:
        """Build a list of random circuits to perform the XEB experiment with.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different numbers of cycles to include in each circuit.

        Returns:
            The list of experiment samples.
        """
        random_circuits = []
        for _, depth in tqdm.contrib.itertools.product(
            range(num_circuits), cycle_depths, desc="Building circuits"
        ):
            num_single_qubit_gate_layers = depth + int(self.two_qubit_gate is not None)
            chosen_single_qubit_gates = self._rng.choice(
                np.asarray(self.single_qubit_gate_set),
                size=(num_single_qubit_gate_layers, self.num_qubits),
            )

            circuit = cirq.Circuit(
                gate.on(qubit)
                for gates_in_layer in chosen_single_qubit_gates
                for gate, qubit in zip(gates_in_layer, self.qubits)
            )

            if self.two_qubit_gate is not None:
                circuit = self._interleave_op(circuit, self.two_qubit_gate(*self.qubits))

            random_circuits.append(
                XEBSample(
                    raw_circuit=circuit + cirq.measure(sorted(circuit.all_qubits())),
                    data={
                        "circuit_depth": len(circuit),
                        "num_cycles": depth,
                        "two_qubit_gate": str(self.two_qubit_gate),
                    },
                )
            )

        return random_circuits

    def _process_probabilities(
        self, samples: Sequence[XEBSample]  # type: ignore[override]
    ) -> pd.DataFrame:
        """Processes the probabilities generated by sampling the circuits into the data structures
        needed for analyzing the results.

        Args:
            samples: The list of samples to process the results from.

        Returns:
            A data frame of the full results needed to analyse the experiment.
        """
        samples = list(samples)
        missing_count = 0
        for sample in samples:
            if sample.probabilities is None:
                missing_count += 1
                samples.remove(sample)
            else:
                sample.sample_probabilities = sample.probabilities
                sample.probabilities = {}
        if missing_count > 0:
            warnings.warn(
                f"{missing_count} sample(s) are missing `probabilities`. "
                "These samples have been omitted."
            )

        for sample in tqdm.auto.tqdm(samples, desc="Evaluating circuits"):
            sample.target_probabilities = self._simulate_sample(sample)

        records = []
        for sample in samples:
            if sample.sample_probabilities is not None and sample.target_probabilities is not None:
                target_probabilities = {
                    f"p({key})": value for key, value in sample.target_probabilities.items()
                }
                sample_probabilities = {
                    f"p^({key})": value for key, value in sample.sample_probabilities.items()
                }
                records.append(
                    {
                        "cycle_depth": sample.data["num_cycles"],
                        "circuit_depth": sample.data["circuit_depth"],
                        **target_probabilities,
                        **sample_probabilities,
                        "sum_p(x)p(x)": sample.sum_target_probs_square(),
                        "sum_p(x)p^(x)": sample.sum_target_cross_sample_probs(),
                    }
                )
        return pd.DataFrame(records)

    def _simulate_sample(self, sample: XEBSample) -> dict[str, float]:
        """Simulates the exact probabilities of measuring all possible bitstrings
        with a given sample.

        Args:
            sample: The sample to simulate.

        Returns:
            A dictionary of the probability of each bitstring.
        """
        sim = cirq.Simulator(seed=self._rng)

        result = sim.simulate(
            cirq.drop_terminal_measurements(sample.circuit),
            qubit_order=sorted(sample.circuit.all_qubits()),
        )
        return {
            f"{i:0{self.num_qubits}b}": np.abs(amp) ** 2
            for i, amp in enumerate(result.final_state_vector)
        }

    ###################
    # Public Methods  #
    ###################
    def analyze_results(self, plot_results: bool = True) -> XEBResults:
        """Analyse the results and calculate the estimated circuit fidelity.

        Args:
            plot_results (optional): Whether to generate the data plots. Defaults to True.

        Returns:
           The final results from the experiment.
        """

        # Fit a linear model for each cycle depth to estimate the circuit fidelity
        records = []
        for depth in set(self.raw_data.cycle_depth):
            df = self.raw_data[self.raw_data.cycle_depth == depth]
            fit = scipy.stats.linregress(
                x=df["sum_p(x)p(x)"] - 1 / 2**self.num_qubits,
                y=df["sum_p(x)p^(x)"] - 1 / 2**self.num_qubits,
            )

            records.append(
                {
                    "cycle_depth": depth,
                    "circuit_fidelity_estimate": fit.slope,
                    "circuit_fidelity_estimate_std": fit.stderr,
                }
            )

        self._circuit_fidelities = pd.DataFrame(records)

        self.circuit_fidelities["log_fidelity_estimate"] = np.log(
            self.circuit_fidelities.circuit_fidelity_estimate
        )

        # Fit a linear model to the depth ~ log(fidelity) to approximate the cycle fidelity
        cycle_fit = scipy.stats.linregress(
            x=self.circuit_fidelities.cycle_depth,
            y=np.log(self.circuit_fidelities.circuit_fidelity_estimate),
        )
        cycle_fidelity_estimate = np.exp(cycle_fit.slope)
        cycle_fidelity_estimate_std = cycle_fidelity_estimate * cycle_fit.stderr

        self._results = XEBResults(
            target="$ ".join(self.targets),
            total_circuits=len(self.samples),
            cycle_fidelity_estimate=cycle_fidelity_estimate,
            cycle_fidelity_estimate_std=cycle_fidelity_estimate_std,
        )

        if plot_results:
            self.plot_results()

        return self.results

    def plot_results(self) -> None:
        """Plot the experiment data and the corresponding fits."""
        plot_1 = sns.lmplot(
            data=self.raw_data,
            x="sum_p(x)p(x)",
            y="sum_p(x)p^(x)",
            hue="cycle_depth",
            palette="dark:r",
            legend="full",
            ci=None,
        )
        sns.move_legend(plot_1, "center right")
        ax_1 = plot_1.axes.item()
        plot_1.tight_layout()
        ax_1.set_xlabel(r"$\sum p(x)^2$", fontsize=15)
        ax_1.set_ylabel(r"$\sum p(x) \hat{p}(x)$", fontsize=15)
        ax_1.set_title(r"Linear fit per cycle depth", fontsize=15)

        plot_2 = sns.lmplot(
            data=self.circuit_fidelities,
            x="cycle_depth",
            y="circuit_fidelity_estimate",
            hue="cycle_depth",
            palette="dark:r",
        )
        ax_2 = plot_2.axes.item()
        plot_2.tight_layout()
        ax_2.set_xlabel(r"Cycle depth", fontsize=15)
        ax_2.set_ylabel(r"Circuit fidelity", fontsize=15)
        ax_2.set_title(r"Exponential decay of circuit fidelity", fontsize=15)

        # Add fit line
        x = np.linspace(
            self.circuit_fidelities.cycle_depth.min(), self.circuit_fidelities.cycle_depth.max()
        )
        y = self.results.cycle_fidelity_estimate**x
        y_p = (self.results.cycle_fidelity_estimate + self.results.cycle_fidelity_estimate_std) ** x
        y_m = (self.results.cycle_fidelity_estimate - self.results.cycle_fidelity_estimate_std) ** x
        ax_2.plot(x, y, color="tab:red", linewidth=2)
        ax_2.fill_between(x, y_m, y_p, alpha=0.2, color="tab:red")



# Copyright 2021 The Cirq Developers
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base experiment class and tools used across all experiments.
"""
from __future__ import annotations

import functools
import math
import pprint
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence
from dataclasses import dataclass, field
from typing import Any, Generic, TypeVar

import cirq
import cirq_superstaq as css
import numpy as np
import pandas as pd
from tqdm.auto import tqdm


@dataclass
class Sample:
    """A sample circuit to use along with any data about the circuit
    that is needed for analysis
    """

    raw_circuit: cirq.Circuit
    """The raw (i.e. pre-compiled) sample circuit."""
    data: dict[str, Any]
    """The corresponding data about the circuit"""
    probabilities: dict[str, float] | None = None
    """The probabilities of the computational basis states"""
    job: css.Job | None = None
    """The superstaq job corresponding to the sample. Defaults to None if no job is
    associated with the sample."""
    compiled_circuit: cirq.Circuit | None = None
    """The compiled circuit. Only used if the circuits are compiled for a specific
    target."""

    @property
    def target(self) -> str:
        """Returns:
        The name of the target that the sample was submitted to.
        """
        if self.job is not None:
            # If there is a job then get the target
            return self.job.target()

        if self.probabilities is not None:
            # If no job, but probabilities have been calculated, infer that a local
            # simulator was used.
            return "Local simulator"

        # Otherwise the experiment hasn't yet been run so there is no target.
        return "No target"

    @property
    def circuit(self) -> cirq.Circuit:
        """Returns:
        The circuit used for the experiment. Defaults to the compiled circuit if available
        and if not returns the raw circuit.
        """
        if self.compiled_circuit is not None:
            return self.compiled_circuit

        return self.raw_circuit


@dataclass(frozen=True)
class BenchmarkingResults(ABC):
    """A dataclass for storing the results of the experiment. Requires subclassing for
    each new experiment type."""

    target: str
    """The target device that was used."""
    total_circuits: int
    """The total number of circuits used in the experiment."""

    experiment_name: str = field(init=False)
    """The name of the experiment."""


ResultsT = TypeVar("ResultsT", bound="BenchmarkingResults")


class BenchmarkingExperiment(ABC, Generic[ResultsT]):
    """Base class for gate benchmarking experiments.

    The interface for implementing these experiments is as follows:

    #. First instantiate the desired experiment object

        .. code::

            experiment = ExampleExperiment(<<args/kwargs>>)

    #. Prepare the circuits and run the experiment on the desired target. This can either be a
       custom simulator or a real device name. For example:

        .. code::

            noise_model = cirq.depolarize(p=0.01, n_qubits=1)
            sim = cirq.DensityMatrixSimulator(noise=noise_model)

            experiment.prepare_experiment(<<args/kwargs>>)
            experiment.run_with_simulator(simulator=sim, <<args/kwargs>>)

    #. Then we analyse the results. If the target was a local simulator this will be available as
       soon as the :code:`run_with_simulator()` method has finished executing. On the other hand
       if a real device was accessed via Superstaq then it may take time for the data to be
       available from the server. The :code:`collect_data()` will return :code:`True` when all
       data has been collected and is ready to be analysed.

       .. code::

            if experiment.collect_data():
                results = experiment.analyze_results(<<args>>)

    #. The final results of the experiment will be stored in the :code:`results` attribute as a
       :class:`BenchmarkingResults` of values, while all the data from the experiment will be
       stored in the :code:`raw_data` attribute as a :class:`~pandas.DataFrame`. Some experiments
       may include additional data attributes for data generated during the analysis.

        .. code::

            results = experiment.results
            data = experiment.raw_data

    Additionally it is possible to pre-compile the experimental circuits for a given device using

    .. code::

        experiment.prepare_experiment(<<args/kwargs>>)
        experiment.compile_circuits(target=<<target_name>>)

    And then to run the experiment using a custom callable function for evaluating the circuits.
    For example this could be a function that uses a connection to a local device.

    .. code::

        experiment.run_with_callable(<<function_name>>)

    When implementing a new experiment, 4 methods need to be implemented:

    #. :meth:`_build_circuits`: Given a number of circuits and an iterable of the different numbers
        of layers to use, return a list of :class:`Sample` objects that need to be sampled during
        the experiment.

    #. :meth:`_process_probabilities`: Take the probability distribution over the
        computational basis resulting from running each circuit and combine the relevant details
        into a :class:`pandas.DataFrame`.

    #. :meth:`analyze_results`: Analyse the data in the :attr:`raw_data` dataframe and return a
        :class:`BenchmarkingResults` object containing the results of the experiment.

    #. :meth:`plot_results`: Produce any relevant plots that are useful for understanding the
        results of the experiment.

    Additionally the :class:`BenchmarkingResults` dataclass needs subclassing to hold the specific
    results of the new experiment.
    """

    def __init__(
        self,
        num_qubits: int,
        *,
        random_seed: int | np.random.Generator | None = None,
        **kwargs: Any,
    ) -> None:
        """Initializes a benchmarking experiment.

        Args:
            num_qubits: The number of qubits used during the experiment. Most subclasses
                will determine this from their other inputs.
            random_seed: An optional seed to use for randomization.
            kwargs: Additional kwargs passed to the Superstaq service object.
        """
        self.qubits = cirq.LineQubit.range(num_qubits)
        """The qubits used in the experiment."""

        self._raw_data: pd.DataFrame | None = None
        "The data generated during the experiment"

        self._results: ResultsT | None = None
        """The attribute to store the results in."""

        self._samples: Sequence[Sample] | None = None
        """The attribute to store the experimental samples in."""

        self._service_kwargs = kwargs
        """Arguments to pass to the Superstaq service for submitting jobs."""

        self._rng = np.random.default_rng(random_seed)

    ##############
    # Properties #
    ##############
    @functools.cached_property
    def _superstaq_service(self) -> css.Service:
        """A Superstaq service to use for compilation and circuit submission."""
        return css.Service(**self._service_kwargs)

    @property
    def num_qubits(self) -> int:
        """Returns:
        The number of qubits used in the experiment
        """
        return len(self.qubits)

    @property
    def raw_data(self) -> pd.DataFrame:
        """The data from the most recently run experiment.

        Raises:
            RuntimeError: If no results are available.
        """
        if self._raw_data is None:
            raise RuntimeError("No data to retrieve. The experiment has not been run.")

        return self._raw_data

    @property
    def results(self) -> ResultsT:
        """The results from the most recently run experiment.

        Raises:
            RuntimeError: If no results are available.
        """
        if self._results is None:
            raise RuntimeError("No results to retrieve. The experiment has not been run.")

        return self._results

    @property
    def samples(self) -> Sequence[Sample]:
        """The samples generated during the experiment.

        Raises:
            RuntimeError: If no samples are available.
        """
        if self._samples is None:
            raise RuntimeError("No samples to retrieve. The experiment has not been run.")

        return self._samples

    @property
    def targets(self) -> frozenset[str]:
        """Returns:
        A set of the unique target that each sample was submitted to
        """
        return frozenset(sample.target for sample in self.samples)

    ###################
    # Private Methods #
    ###################
    @abstractmethod
    def _build_circuits(
        self,
        num_circuits: int,
        cycle_depths: Iterable[int],
    ) -> Sequence[Sample]:
        """Build a list of circuits required for the experiment. These circuits are stored in
        :class:`Sample` objects along with any additional data that is needed during the analysis.

        Args:
            num_circuits: Number of circuits to generate.
            cycle_depths: An iterable of the different cycle depths to use during the experiment.

        Returns:
           The list of experiment samples.
        """

    @staticmethod
    def _interleave_op(
        circuit: cirq.Circuit, operation: cirq.Operation, include_final: bool = False
    ) -> cirq.Circuit:
        """Interleave a given operation into a circuit.

        Args:
            circuit: The original circuit.
            operation: The operation to interleave.
            include_final: If True then the interleaving gate is also appended to
                the end of the circuit.

        Returns:
            A copy of the original circuit with the provided gate interleaved.
        """
        operation = operation.with_tags("no_compile")
        interleaved_circuit = circuit.copy()
        interleaved_circuit.batch_insert(
            [(k, operation) for k in range(len(circuit) - int(not include_final), 0, -1)]
        )
        return interleaved_circuit

    def _process_device_counts(self, counts: dict[str, int]) -> dict[str, float]:
        """Process the counts returned by the server into a dictionary of probabilities.

        Args:
            counts: A dictionary of the observed counts for each state in the computational basis.

        Returns:
            A dictionary of the probability of each state in the computational basis.
        """
        total = sum(counts.values())

        probabilities = {
            format(idx, f"0{self.num_qubits}b"): 0.0 for idx in range(2**self.num_qubits)
        }

        for key, count in counts.items():
            probabilities[key] = count / total

        return probabilities

    @abstractmethod
    def _process_probabilities(self, samples: Sequence[Sample]) -> pd.DataFrame:
        """Processes the probabilities generated by sampling the circuits into a data frame
        needed for analyzing the results.

        Args:
            samples: The list of samples to process the results from.

        Returns:
            A data frame of the full results needed to analyse the experiment.
        """

    def _retrieve_jobs(self) -> dict[str, str]:
        """Retrieve the jobs from the server.

        Returns:
            A dictionary of the statuses of any jobs that have not been successfully completed.
        """
        # If no jobs then return empty dictionary
        if not [sample for sample in self.samples if sample.job is not None]:
            return {}

        statuses = {}
        waiting_samples = [sample for sample in self.samples if sample.probabilities is None]
        for sample in tqdm(waiting_samples, "Retrieving jobs"):
            if sample.job is None:
                continue
            if (
                job_status := sample.job.status()
            ) in css.job.Job.NON_TERMINAL_STATES + css.job.Job.UNSUCCESSFUL_STATES:
                statuses[sample.job.job_id()] = job_status
            else:
                sample.probabilities = self._process_device_counts(sample.job.counts(0))

        return statuses

    def _has_raw_data(self) -> None:
        """Checks if any of the samples already have probabilities stored. If so raises a runtime
        error to prevent them from being overwritten.

        To be used within all `run` methods to prevent data being overwritten

        Raises:
            RuntimeError: If any samples already have probabilities stored.
        """
        if any(sample.probabilities is not None for sample in self.samples):
            raise RuntimeError(
                "Some samples have already been run. Re-running the experiment will"
                "overwrite these results. If this is the desired behaviour use `overwrite=True`"
            )

    def _sample_statuses(self) -> list[str | None]:
        """Returns:
        The statuses of the jobs associated with each sample. If no job is associated
        with a sample then :code:`None` is listed instead.
        """
        statuses: list[str | None] = []
        for sample in self.samples:
            if sample.job is None:
                statuses.append(None)
            else:
                statuses.append(sample.job.status())
        return statuses

    def _validate_circuits(self) -> None:
        """Checks that all circuits contain a terminal measurement of all qubits."""
        for sample in self.samples:
            if not sample.circuit.has_measurements():
                raise ValueError("QCVV experiment circuits must contain measurements.")
            if not sample.circuit.are_all_measurements_terminal():
                raise ValueError("QCVV experiment circuits can only contain terminal measurements.")
            if not sorted(sample.circuit[-1].qubits) == sorted(self.qubits):
                raise ValueError(
                    "The terminal measurement in QCVV experiment circuits must measure all qubits."
                )

    ###################
    # Public Methods  #
    ###################
    @abstractmethod
    def analyze_results(self, plot_results: bool = True) -> ResultsT:
        """Perform the experiment analysis and store the results in the `results` attribute.

        Args:
            plot_results: Whether to generate plots of the results. Defaults to False.

        Returns:
            A named tuple of the final results from the experiment.
        """

    def collect_data(self, force: bool = False) -> bool:
        """Collect the data from the samples and process it into the :attr:`raw_data` attribute.

        If the data is successfully stored in the :attr:`raw_data` attribute then the function will
        return :code:`True`.

        If either not all jobs submitted to the server have completed, or not all samples have
        probability results then no data will be saved in :attr:`raw_data` and the function will
        return :code:`False`. This check can be overridden with :code:`force=True` in which case
        only the samples which have probability results will be used to generate the results
        dataframe.

        Args:
            force: Whether to override the check that all data is present. Defaults to False.

        Raises:
            RuntimeError: If :code:`force=True` but there are no samples with any data.
            RuntimeError: If the experiment has not yet been run.

        Returns:
            Whether the results dataframe has been successfully created.
        """
        if not self._samples:
            raise RuntimeError("The experiment has not yet ben run.")

        # Retrieve jobs from server (if needed)
        outstanding_statuses = self._retrieve_jobs()
        if outstanding_statuses:
            print(
                "Not all circuits have been sampled. "
                "Please wait and try again.\n"
                f"Outstanding Superstaq jobs:\n{pprint.pformat(outstanding_statuses)}"
            )
            if not force:
                return False

        completed_samples = [sample for sample in self.samples if sample.probabilities is not None]

        if not len(completed_samples) == len(self.samples):
            print("Some samples do not have probability results.")
            if not force:
                return False

        if len(completed_samples) == 0:
            raise RuntimeError("Cannot force data collection when there are no completed samples.")

        self._raw_data = self._process_probabilities(completed_samples)
        return True

    def compile_circuits(self, target: str, **kwargs: Any) -> None:
        """Compiles the experiment circuits for the given device. Useful if the samples
        are not going to be run via Superstaq.

        Args:
            target: The device to compile to.
            kwargs: Additional desired compile options.
        """
        compiled_circuits = self._superstaq_service.compile(
            [sample.circuit for sample in self.samples], target=target, **kwargs
        ).circuits

        for k, sample in enumerate(self.samples):
            sample.compiled_circuit = compiled_circuits[k]  # type: ignore[assignment]

    @abstractmethod
    def plot_results(self) -> None:
        """Plot the results of the experiment"""

    def prepare_experiment(
        self, num_circuits: int, cycle_depths: Iterable[int], overwrite: bool = False
    ) -> None:
        """Prepares the circuits needed for the experiment

        Args:
            num_circuits: Number of circuits to run.
            cycle_depths: An iterable of the different layer depths to use during the experiment.
            overwrite: Whether to force an experiment run even if there is existing data that would
                be over written in the process. Defaults to False.

        Raises:
            RuntimeError: If the experiment has already been run once and the `overwrite` argument
                is not True
            ValueError: If any of the cycle depths provided negative or zero.
        """
        if self._samples is not None and not overwrite:
            raise RuntimeError(
                "This experiment already has existing data which would be overwritten by "
                "rerunning the experiment. If this is the desired behavior set `overwrite=True`"
            )

        if any(depth <= 0 for depth in cycle_depths):
            raise ValueError("The `cycle_depths` iterator can only include positive values.")

        self._samples = self._build_circuits(num_circuits, cycle_depths)
        self._validate_circuits()

    def run_on_device(
        self,
        target: str,
        repetitions: int = 10_000,
        method: str | None = None,
        overwrite: bool = False,
        **target_options: Any,
    ) -> css.Job:
        """Submit the circuit samples to the desired target device and store the resulting
        probabilities.

        The set of circuits is partitioned as necessary to not exceed the maximum circuits that can
        be submitted to the given target device. The function then waits for the jobs to complete
        before saving the resulting probability distributions.

        Args:
            target: The name of a Superstaq target.
            repetitions: The number of shots to sample. Defaults to 10,000.
            method: Optional method to use on the Superstaq device. Defaults to None corresponding
                to normal running.
            target_options: Optional configuration dictionary passed when submitting the job.
            overwrite: Whether to force an experiment run even if there is existing data that would
                be over written in the process. Defaults to False.

        Return:
            The superstaq job containing all the circuits submitted as part of the experiment.
        """
        if not overwrite:
            self._has_raw_data()

        experiment_job = self._superstaq_service.create_job(
            [sample.circuit for sample in self.samples],
            target=target,
            method=method,
            repetitions=repetitions,
            **target_options,
        )
        compiled_circuits = experiment_job.compiled_circuits()

        for k, sample in enumerate(self.samples):
            sample.job = experiment_job[k]
            sample.compiled_circuit = compiled_circuits[k]

        return experiment_job

    def run_with_simulator(
        self,
        simulator: cirq.Sampler | None = None,
        repetitions: int = 10_000,
        overwrite: bool = False,
    ) -> None:
        """Use the local simulator to sample the circuits and store the resulting probabilities.

        Args:
            simulator: A local :class:`~cirq.Sampler` to use. If None then the default
                :class:`cirq.Simulator` simulator is used. Defaults to None.
            repetitions: The number of shots to sample. Defaults to 10,000.
            overwrite: Whether to force an experiment run even if there is existing data that would
                be over written in the process. Defaults to False.
        """
        if not overwrite:
            self._has_raw_data()

        if simulator is None:
            simulator = cirq.Simulator(seed=self._rng)

        for sample in tqdm(self.samples, desc="Simulating circuits"):
            result = simulator.run(sample.circuit, repetitions=repetitions)
            hist = result.histogram(key=cirq.measurement_key_name(sample.circuit))
            sample.probabilities = {
                f"{i:0{self.num_qubits}b}": 0.0 for i in range(2**self.num_qubits)
            }  # Set all probabilities to zero
            for val, count in hist.items():
                # Add in results from the histogram
                sample.probabilities[f"{val:0{self.num_qubits}b}"] = count / repetitions

    def run_with_callable(
        self,
        circuit_eval_func: Callable[[cirq.Circuit], dict[str, float]],
        overwrite: bool = False,
        **kwargs: Any,
    ) -> None:
        """Evaluates the probabilities for each circuit using a user provided callable function.
        This function should take a circuit as input and return a dictionary of probabilities for
        each bitstring (including states with zero probability).

        Args:
            circuit_eval_func: The custom function to use when evaluating circuit probabilities.
            overwrite: Whether to force an experiment run even if there is existing data that would
                be over written in the process. Defaults to False.
            kwargs: Additional arguments to pass to the custom function.

        Raises:
            RuntimeError: If the returned probabilities dictionary keys is missing include
                an incorrect number of bits.
            RuntimeError: If the returned probabilities dictionary values do not sum to 1.0.
        """
        if not overwrite:
            self._has_raw_data()
        for sample in tqdm(self.samples, desc="Running circuits"):
            probability = circuit_eval_func(sample.circuit, **kwargs)
            if not all(len(key) == self.num_qubits for key in probability.keys()):
                raise RuntimeError("Returned probabilities include an incorrect number of bits.")
            if not math.isclose(sum(probability.values()), 1.0):
                raise RuntimeError("Returned probabilities do not sum to 1.0.")

            for k in range(2**self.num_qubits):
                if (bitstring := format(k, f"0{self.num_qubits}b")) not in probability:
                    probability[bitstring] = 0.0

            sample.probabilities = probability



"""A toolkit of QCVV routines."""

from .base_experiment import BenchmarkingExperiment, BenchmarkingResults, Sample
from .irb import IRB, IRBResults
from .xeb import XEB, XEBResults, XEBSample

__all__ = [
    "BenchmarkingExperiment",
    "BenchmarkingResults",
    "Sample",
    "IRB",
    "IRBResults",
    "XEB",
    "XEBResults",
    "XEBSample",
]
